import os

import pytest

from d8s_file_system import atomic_write, directory_create, file_read, directory_delete

TEST_DIRECTORY_PATH = './test_files'
NON_EXISTENT_FILE_PATH = './foo'
TEST_FILE_NAME = 'a'
EXISTING_FILE_PATH = os.path.join(TEST_DIRECTORY_PATH, TEST_FILE_NAME)


def test_atomic_write_docs_1():
    directory_create(TEST_DIRECTORY_PATH)
    FILE_CONTENTS = 'foo'

    with atomic_write(EXISTING_FILE_PATH) as f:
        f.write(FILE_CONTENTS)
    assert file_read(EXISTING_FILE_PATH) == FILE_CONTENTS

    with atomic_write(EXISTING_FILE_PATH) as f:
        f.write(FILE_CONTENTS)
    assert file_read(EXISTING_FILE_PATH) == FILE_CONTENTS

    # if we try to write to an existing file with `overwrite=False`, we will get an error
    with pytest.raises(FileExistsError):
        with atomic_write(EXISTING_FILE_PATH, overwrite=False) as f:
            f.write(FILE_CONTENTS)

    directory_delete(TEST_DIRECTORY_PATH)



import os
import tempfile

import pytest

from d8s_file_system import directory_create, directory_delete, directory_file_names
from d8s_file_system import (
    is_file,
    file_read,
    file_read_bytes,
    file_write,
    file_append,
    file_move,
    file_copy,
    file_delete,
    file_owner_name,
    file_change_owner,
    file_ssdeep,
    file_md5,
    file_sha1,
    file_sha256,
    file_sha512,
    file_name_escape,
    file_name,
    file_name_windows,
    file_name_unix,
    file_size,
    file_directory,
    file_details,
    file_exists,
    file_is_readable,
    file_is_writable,
    file_is_executable,
    file_contains,
    file_search,
    file_name_matches,
    temp_file_create,
)
from d8s_file_system.files import _file_active_action, _file_action

NON_EXISTENT_FILE_PATH = './foo'
TEST_DIRECTORY_PATH = './test_files'
TEST_FILE_CONTENTS = 'a'
TEST_FILE_NAME = 'a'
EXISTING_FILE_PATH = os.path.join(TEST_DIRECTORY_PATH, TEST_FILE_NAME)


@pytest.fixture(autouse=True)
def clear_testing_directory():
    """This function is run after every test."""
    directory_delete(TEST_DIRECTORY_PATH)
    directory_create(TEST_DIRECTORY_PATH)
    file_write(EXISTING_FILE_PATH, TEST_FILE_CONTENTS)


def setup_module():
    """This function is run before all of the tests in this file are run."""
    directory_create(TEST_DIRECTORY_PATH)


def teardown_module():
    """This function is run after all of the tests in this file are run."""
    directory_delete(TEST_DIRECTORY_PATH)


def test_file_append_1():
    contents = file_read(EXISTING_FILE_PATH)
    assert contents == TEST_FILE_CONTENTS
    file_append(EXISTING_FILE_PATH, 'bc')
    contents = file_read(EXISTING_FILE_PATH)
    assert contents == 'abc'


def test_file_copy_1():
    result = file_copy(EXISTING_FILE_PATH, os.path.join(TEST_DIRECTORY_PATH, 'b'))
    assert directory_file_names(TEST_DIRECTORY_PATH) == ['a', 'b']

    # TODO: test the preserve_metadata kwarg (this requires a function to get file metadata)


def test_file_delete_1():
    assert directory_file_names(TEST_DIRECTORY_PATH) == ['a']
    result = file_delete(EXISTING_FILE_PATH)
    assert directory_file_names(TEST_DIRECTORY_PATH) == []


def test_file_write_1():
    contents = file_read(EXISTING_FILE_PATH)
    assert contents == TEST_FILE_CONTENTS

    file_write(EXISTING_FILE_PATH, 'foo bar')
    contents = file_read(EXISTING_FILE_PATH)
    assert contents == 'foo bar'

    file_write(EXISTING_FILE_PATH, b'abc')
    contents = file_read(EXISTING_FILE_PATH)
    assert contents == 'abc'

    file_write(EXISTING_FILE_PATH, [1, 2, 3])
    contents = file_read(EXISTING_FILE_PATH)
    assert contents == '[1, 2, 3]'


def test_file_move_1():
    assert directory_file_names(TEST_DIRECTORY_PATH) == ['a']
    file_move(EXISTING_FILE_PATH, os.path.join(TEST_DIRECTORY_PATH, 'b'))
    assert directory_file_names(TEST_DIRECTORY_PATH) == ['b']


def test_file_contains_docs_1():
    assert file_contains(EXISTING_FILE_PATH, 'a') == True
    assert file_contains(EXISTING_FILE_PATH, 'b') == False
    assert file_contains(EXISTING_FILE_PATH, '[abc]', pattern_is_regex=True) == True
    # assert file_contains(NON_EXISTENT_FILE_PATH, 'a') == 'fill'  # [Errno 2] No such file or directory


def test_file_details_docs_1():
    assert file_details(EXISTING_FILE_PATH) == {
        'md5': '0cc175b9c0f1b6a831c399e269772661',
        'sha1': '86f7e437faa5a7fce15d1ddcb9eaeaea377667b8',
        'sha256': 'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb',
        'size': 1,
        'ssdeep': '3:E:E',
    }
    with pytest.raises(FileNotFoundError):
        file_details(NON_EXISTENT_FILE_PATH)


def test_file_exists_docs_1():
    assert file_exists(EXISTING_FILE_PATH) == True
    assert file_exists(NON_EXISTENT_FILE_PATH) == False


def test_file_is_executable_docs_1():
    assert file_is_executable(EXISTING_FILE_PATH) == False
    assert file_is_executable(NON_EXISTENT_FILE_PATH) == False


def test_file_is_readable_docs_1():
    assert file_is_readable(EXISTING_FILE_PATH) == True
    assert file_is_readable(NON_EXISTENT_FILE_PATH) == False


def test_file_is_writable_docs_1():
    assert file_is_writable(EXISTING_FILE_PATH) == True
    assert file_is_writable(NON_EXISTENT_FILE_PATH) == False


def test_file_md5_docs_1():
    assert file_md5(EXISTING_FILE_PATH) == '0cc175b9c0f1b6a831c399e269772661'
    with pytest.raises(FileNotFoundError):
        file_md5(NON_EXISTENT_FILE_PATH)


def test_file_name_docs_1():
    assert file_name(EXISTING_FILE_PATH) == 'a'
    assert file_name(NON_EXISTENT_FILE_PATH) == 'foo'


def test_file_name_escape_docs_1():
    assert file_name_escape('a') == 'a'


def test_file_name_matches_docs_1():
    assert file_name_matches('a', 'a')
    assert file_name_matches('a', '?')
    assert file_name_matches('a', 'a*')
    assert not file_name_matches('a', 'a?')
    assert not file_name_matches('a', 'b')
    assert file_name_matches('a', '[abc]')
    assert not file_name_matches('a', '[bc]')

    path = '/Users/me/code/democritus/foo/democritus_sample_docs.json'
    assert file_name_matches(path, 'democritus_*_docs.json')


def test_file_name_unix_docs_1():
    assert file_name_unix('/Users/bob/documents/Summer2018.pdf') == 'Summer2018.pdf'


def test_file_name_windows_docs_1():
    assert file_name_windows('C:\\Documents\\Newsletters\\Summer2018.pdf') == 'Summer2018.pdf'

    # def test_file_owner_name_docs_1():
    with pytest.raises(FileNotFoundError):
        file_owner_name(NON_EXISTENT_FILE_PATH)


def test_file_read_docs_1():
    assert file_read(EXISTING_FILE_PATH) == 'a'
    with pytest.raises(FileNotFoundError):
        file_read(NON_EXISTENT_FILE_PATH)


def test_file_read_bytes_docs_1():
    assert file_read_bytes(EXISTING_FILE_PATH) == b'a'
    with pytest.raises(FileNotFoundError):
        file_read_bytes(NON_EXISTENT_FILE_PATH)


def test_file_search_docs_1():
    assert file_search(EXISTING_FILE_PATH, 'a') == ['a']
    assert file_search(EXISTING_FILE_PATH, 'b') == []
    assert file_search(EXISTING_FILE_PATH, '[abc]', pattern_is_regex=True) == ['a']
    with pytest.raises(FileNotFoundError):
        file_search(NON_EXISTENT_FILE_PATH, 'a')


def test_file_sha1_docs_1():
    assert file_sha1(EXISTING_FILE_PATH) == '86f7e437faa5a7fce15d1ddcb9eaeaea377667b8'
    with pytest.raises(FileNotFoundError):
        file_sha1(NON_EXISTENT_FILE_PATH)


def test_file_sha256_docs_1():
    assert file_sha256(EXISTING_FILE_PATH) == 'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
    with pytest.raises(FileNotFoundError):
        file_sha256(NON_EXISTENT_FILE_PATH)


def test_file_sha512_docs_1():
    assert (
        file_sha512(EXISTING_FILE_PATH)
        == '1f40fc92da241694750979ee6cf582f2d5d7d28e18335de05abc54d0560e0f5302860c652bf08d560252aa5e74210546f369fbbbce8c12cfc7957b2652fe9a75'
    )
    with pytest.raises(FileNotFoundError):
        file_sha512(NON_EXISTENT_FILE_PATH)


def test_file_size_docs_1():
    assert file_size(EXISTING_FILE_PATH) == 1
    with pytest.raises(FileNotFoundError):
        file_size(NON_EXISTENT_FILE_PATH)


def test_file_ssdeep_docs_1():
    assert file_ssdeep(EXISTING_FILE_PATH) == '3:E:E'
    with pytest.raises(FileNotFoundError):
        file_ssdeep(NON_EXISTENT_FILE_PATH)


def test_is_file_docs_1():
    assert is_file(EXISTING_FILE_PATH) == True
    assert is_file(NON_EXISTENT_FILE_PATH) == False


def test_temp_file_create_docs_1():
    temp_file = temp_file_create()
    assert temp_file
    temp_file.close()



import contextlib
import os

from atomicwrites import atomic_write as atomic_write_, AtomicWriter


class AtomicWriterPerms(AtomicWriter):
    """This class wraps the AtomicWriter from the atomicwrites package to update the file permissions after the file is created.

    This snippet was taken from/inspired by the code here: https://github.com/OCR-D/core/pull/625."""

    def get_fileobject(self, **kwargs):
        f = super().get_fileobject(**kwargs)
        try:
            mode = os.stat(self._path).st_mode
        except FileNotFoundError:
            # Creating a new file, emulate what os.open() does
            mask = os.umask(0)
            os.umask(mask)
            mode = 0o664 & ~mask
        fd = f.fileno()
        os.fchmod(fd, mode)
        return f


@contextlib.contextmanager
def atomic_write(fpath, *, overwrite: bool = True, **cls_kwargs):
    with atomic_write_(fpath, writer_cls=AtomicWriterPerms, overwrite=overwrite, **cls_kwargs) as f:
        yield f



import os
import tempfile

import pytest

from d8s_file_system import (
    is_directory,
    directory_exists,
    directory_file_names,
    directory_file_paths,
    directory_copy,
    directory_delete,
    directory_create,
    directory_disk_usage,
    directory_disk_free_space,
    directory_disk_used_space,
    directory_disk_total_space,
    home_directory,
    home_directory_join,
    directory_move,
    directory_files_details,
    directory_files_read,
    directory_subdirectory_names,
    directory_files_containing,
    directory_file_paths_matching,
    directory_file_names_matching,
    directory_read_files_with_path_matching,
    temp_dir_create,
)
from d8s_file_system import file_write

NON_EXISTENT_DIRECTORY_PATH = './foo'
EXISTING_DIRECTORY_PATH = './test_directories'


@pytest.fixture(autouse=True)
def clear_testing_directory():
    """This function is run after every test."""
    if directory_exists(EXISTING_DIRECTORY_PATH):
        directory_delete(EXISTING_DIRECTORY_PATH)
    if directory_exists(NON_EXISTENT_DIRECTORY_PATH):
        directory_delete(NON_EXISTENT_DIRECTORY_PATH)
    directory_create(EXISTING_DIRECTORY_PATH)
    file_write(os.path.join(EXISTING_DIRECTORY_PATH, 'a'), 'a')
    file_write(os.path.join(EXISTING_DIRECTORY_PATH, 'b'), 'b')
    file_write(os.path.join(EXISTING_DIRECTORY_PATH, 'c'), 'c')


def setup_module():
    """This function is run before all of the tests in this file are run."""
    directory_create(EXISTING_DIRECTORY_PATH)


def teardown_module():
    """This function is run after all of the tests in this file are run."""
    directory_delete(EXISTING_DIRECTORY_PATH)


def test_directory_create_docs_1():
    directory_create(os.path.join(EXISTING_DIRECTORY_PATH, 'foo'))
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == ['foo']

    # creating the same directory again works (it does not raise an error)
    directory_create(os.path.join(EXISTING_DIRECTORY_PATH, 'foo'))
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == ['foo']

    # creating multiple layers of directories works too
    directory_create(os.path.join(EXISTING_DIRECTORY_PATH, 'foo', 'subfoo'))
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == ['foo']
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH) == ['foo', 'subfoo']


def test_directory_copy_docs_1():
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == []
    new_directory_path = os.path.join(EXISTING_DIRECTORY_PATH, 'foo')
    directory_create(new_directory_path)
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == ['foo']
    file_write(os.path.join(new_directory_path, 'a'), 'a')

    copy_directory_path = os.path.join(EXISTING_DIRECTORY_PATH, 'bar')
    directory_copy(new_directory_path, copy_directory_path)
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == ['foo', 'bar']
    assert directory_file_names(new_directory_path) == ['a']
    assert directory_file_names(copy_directory_path) == ['a']

    # copying into a directory which already exists raises an error
    with pytest.raises(FileExistsError):
        directory_copy(new_directory_path, copy_directory_path)


def test_directory_copy_docs_recursive_copy():
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == []
    new_directory_path = os.path.join(EXISTING_DIRECTORY_PATH, 'foo', 'subfoo')
    directory_create(new_directory_path)
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH) == ['foo', 'subfoo']
    file_write(os.path.join(new_directory_path, 'a'), 'a')

    # copying a directory copies everything inside that directory
    src_path = os.path.join(EXISTING_DIRECTORY_PATH, 'foo')
    dst_path = os.path.join(EXISTING_DIRECTORY_PATH, 'bar')
    deep_dst_path = os.path.join(dst_path, 'subfoo')
    directory_copy(src_path, dst_path)
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH) == ['foo', 'bar', 'subfoo', 'subfoo']
    assert directory_file_names(deep_dst_path) == ['a']


def test_directory_move_docs_1():
    assert directory_exists(EXISTING_DIRECTORY_PATH)
    assert not directory_exists(NON_EXISTENT_DIRECTORY_PATH)
    directory_move(EXISTING_DIRECTORY_PATH, NON_EXISTENT_DIRECTORY_PATH)
    assert not directory_exists(EXISTING_DIRECTORY_PATH)
    assert directory_exists(NON_EXISTENT_DIRECTORY_PATH)


def test_directory_delete_docs_1():
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == []
    new_directory_path = os.path.join(EXISTING_DIRECTORY_PATH, 'foo')
    directory_create(new_directory_path)
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == ['foo']
    directory_delete(new_directory_path)
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == []

    with pytest.raises(FileNotFoundError):
        directory_delete(NON_EXISTENT_DIRECTORY_PATH)


def test_home_directory_docs_1():
    result = home_directory()
    assert result.startswith('/Users/')

    result = home_directory_join(EXISTING_DIRECTORY_PATH)
    assert result.startswith('/Users/')
    assert result.endswith('test_directories')


def test_directory_files_containing_docs_1():
    result = directory_files_containing(EXISTING_DIRECTORY_PATH, 'a')
    assert result == {'./test_directories/a': ['a']}

    result = directory_files_containing(EXISTING_DIRECTORY_PATH, 'foo bar')
    assert result == {}

    result = directory_files_containing(EXISTING_DIRECTORY_PATH, '[abc]', pattern_is_regex=True)
    assert result == {'./test_directories/a': ['a'], './test_directories/c': ['c'], './test_directories/b': ['b']}

    result = directory_files_containing(NON_EXISTENT_DIRECTORY_PATH, 'a')
    assert result == {}


def test_directory_read_files_with_path_matching_docs_1():
    result = directory_read_files_with_path_matching(EXISTING_DIRECTORY_PATH, 'a')
    assert list(result) == [('./test_directories/a', 'a')]

    result = directory_read_files_with_path_matching(EXISTING_DIRECTORY_PATH, '*a')
    assert list(result) == [('./test_directories/a', 'a')]

    result = directory_read_files_with_path_matching(EXISTING_DIRECTORY_PATH, 'foo bar')
    assert list(result) == []

    result = directory_read_files_with_path_matching(EXISTING_DIRECTORY_PATH, '*[abc]')
    assert list(result) == [('./test_directories/a', 'a'), ('./test_directories/c', 'c'), ('./test_directories/b', 'b')]

    result = directory_read_files_with_path_matching(NON_EXISTENT_DIRECTORY_PATH, 'a')
    assert list(result) == []


def test_directory_disk_free_space_docs_1():
    assert isinstance(directory_disk_free_space(EXISTING_DIRECTORY_PATH), int)
    with pytest.raises(FileNotFoundError):
        directory_disk_free_space(NON_EXISTENT_DIRECTORY_PATH)


def test_directory_disk_total_space_docs_1():
    assert isinstance(directory_disk_total_space(EXISTING_DIRECTORY_PATH), int)
    with pytest.raises(FileNotFoundError):
        directory_disk_total_space(NON_EXISTENT_DIRECTORY_PATH)


def test_directory_disk_used_space_docs_1():
    assert isinstance(directory_disk_used_space(EXISTING_DIRECTORY_PATH), int)
    with pytest.raises(FileNotFoundError):
        directory_disk_used_space(NON_EXISTENT_DIRECTORY_PATH)


def test_directory_exists_docs_1():
    assert directory_exists(EXISTING_DIRECTORY_PATH) == True
    assert directory_exists(NON_EXISTENT_DIRECTORY_PATH) == False


def test_directory_file_names_docs_1():
    assert directory_file_names(EXISTING_DIRECTORY_PATH) == ['a', 'c', 'b']
    assert directory_file_names(NON_EXISTENT_DIRECTORY_PATH) == []


def test_directory_file_paths_matching_docs_1():
    assert directory_file_paths_matching(EXISTING_DIRECTORY_PATH, '*a') == ['./test_directories/a']
    assert directory_file_paths_matching(EXISTING_DIRECTORY_PATH, '*b') == ['./test_directories/b']
    assert directory_file_paths_matching(NON_EXISTENT_DIRECTORY_PATH, '*b') == []


def test_directory_file_names_matching_docs_1():
    assert directory_file_names_matching(EXISTING_DIRECTORY_PATH, 'a') == ['a']
    assert directory_file_names_matching(EXISTING_DIRECTORY_PATH, 'b') == ['b']
    assert directory_file_names_matching(NON_EXISTENT_DIRECTORY_PATH, 'b') == []


def test_directory_file_paths_docs_1():
    assert directory_file_paths(EXISTING_DIRECTORY_PATH) == [
        './test_directories/a',
        './test_directories/c',
        './test_directories/b',
    ]
    assert directory_file_paths(EXISTING_DIRECTORY_PATH, recursive=False) == [
        './test_directories/a',
        './test_directories/c',
        './test_directories/b',
    ]
    assert directory_file_paths(NON_EXISTENT_DIRECTORY_PATH) == []


def test_directory_files_details_docs_1():
    assert directory_files_details(EXISTING_DIRECTORY_PATH) == {
        './test_directories/a': {
            'md5': '0cc175b9c0f1b6a831c399e269772661',
            'sha1': '86f7e437faa5a7fce15d1ddcb9eaeaea377667b8',
            'sha256': 'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb',
            'ssdeep': '3:E:E',
            'size': 1,
        },
        './test_directories/c': {
            'md5': '4a8a08f09d37b73795649038408b5f33',
            'sha1': '84a516841ba77a5b4648de2cd0dfcb30ea46dbb4',
            'sha256': '2e7d2c03a9507ae265ecf5b5356885a53393a2029d241394997265a1a25aefc6',
            'ssdeep': '3:G:G',
            'size': 1,
        },
        './test_directories/b': {
            'md5': '92eb5ffee6ae2fec3ad71c777531578f',
            'sha1': 'e9d71f5ee7c92d6dc9e92ffdad17b8bd49418f98',
            'sha256': '3e23e8160039594a33894f6564e1b1348bbd7a0088d42c4acb73eeaed59c009d',
            'ssdeep': '3:H:H',
            'size': 1,
        },
    }
    assert directory_files_details(NON_EXISTENT_DIRECTORY_PATH) == {}


def test_directory_files_read_docs_1():
    assert tuple(directory_files_read(EXISTING_DIRECTORY_PATH)) == (
        ('./test_directories/a', 'a'),
        ('./test_directories/c', 'c'),
        ('./test_directories/b', 'b'),
    )
    assert tuple(directory_files_read(NON_EXISTENT_DIRECTORY_PATH)) == ()


def test_directory_subdirectory_names_docs_1():
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH) == []

    new_directory_path = os.path.join(EXISTING_DIRECTORY_PATH, 'foo', 'subfoo')
    directory_create(new_directory_path)
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH) == ['foo', 'subfoo']
    assert directory_subdirectory_names(EXISTING_DIRECTORY_PATH, recursive=False) == ['foo']
    assert directory_subdirectory_names(NON_EXISTENT_DIRECTORY_PATH) == []


def test_is_directory_docs_1():
    assert is_directory(EXISTING_DIRECTORY_PATH) == True
    assert is_directory(NON_EXISTENT_DIRECTORY_PATH) == False


def test_temp_dir_create_docs_1():
    temp_dir = temp_dir_create()
    assert isinstance(temp_dir, tempfile.TemporaryDirectory)
    temp_dir.cleanup()



import os
import shutil
import tempfile
from typing import Any, Dict, Iterable, Tuple, Union, List, Dict

# TODO: test and standardize what happens if these functions are given a directory which does not exist
# TODO: may want to convert some of these functions to use this library: https://pypi.org/project/path.py/
# TODO: Write a function to change the permissions on a file (approx. chmod - https://www.tutorialspoint.com/unix/unix-file-permission.htm)
# TODO: add decorators to these functions (e.g. map_first_arg decorator to file functions)


def is_directory(path: str) -> bool:
    """Determine if the given path is a directory."""
    return os.path.isdir(path) or False


def directory_exists(directory_path: str) -> bool:
    """Check if the directory exists."""
    return os.path.isdir(directory_path) or False


def directory_file_names(directory_path: str, *, recursive: bool = True) -> List[str]:
    """List files at the given directory_path."""
    directory_files = []
    for path, dirs, files in os.walk(directory_path):
        directory_files.extend(files)
        if not recursive:
            break
    return directory_files


def directory_file_paths(directory_path: str, *, recursive: bool = True) -> List[str]:
    """List the file paths at the given directory_path."""
    file_paths = []
    for path, dirs, files in os.walk(directory_path):
        file_paths.extend([os.path.join(path, file) for file in files])
        if not recursive:
            break
    return file_paths


def directory_copy(src_path: str, dst_path: str):
    """Copy the directory from the src_path to the destination path."""
    # TODO: add option to overwrite existing directory
    shutil.copytree(src_path, dst_path)


def directory_delete(directory_path: str):
    """Delete the given directory."""
    shutil.rmtree(directory_path)


def directory_create(directory_path: str, mode=0o777):
    """Create a directory."""
    if not directory_exists(directory_path):
        os.makedirs(directory_path, mode=mode)
    else:
        print(f'Output directory ({directory_path}) already exists')


# TODO: add return types for some of the functions below
def directory_disk_usage(directory_path: str):
    """Return the disk usage for the given directory."""
    return shutil.disk_usage(directory_path)


def directory_disk_free_space(directory_path: str):
    """Return the free space in the given directory."""
    return directory_disk_usage(directory_path).free


def directory_disk_used_space(directory_path: str):
    """Return the used space in the given directory."""
    return directory_disk_usage(directory_path).used


def directory_disk_total_space(directory_path: str):
    """Return the total space in the given directory."""
    return directory_disk_usage(directory_path).total


def home_directory() -> str:
    return os.path.expanduser("~")


def home_directory_join(path: str) -> str:
    """Join the given path with the home directory."""
    return os.path.join(home_directory(), path)


def directory_move(src_path: str, dst_path: str):
    """Move the directory from the src_path to the dst_path."""
    shutil.move(src_path, dst_path)


def directory_files_details(directory_path: str, *, recursive: bool = True) -> Dict[str, Dict[str, Union[str, int]]]:
    """Return the file details for each file in the directory at the given path."""
    from .files import file_details, file_name

    file_paths = directory_file_paths(directory_path, recursive=recursive)
    file_details_dict = {path: file_details(path) for path in file_paths}
    return file_details_dict


def directory_files_read(directory_path: str, *, recursive: bool = True) -> Iterable[Tuple[str, str]]:
    """Read all files in the directory_path."""
    from .files import file_read, file_name

    file_paths = directory_file_paths(directory_path, recursive=recursive)
    for path in file_paths:
        yield path, file_read(path)


def directory_subdirectory_names(directory_path: str, *, recursive: bool = True) -> List[str]:
    """List the names of all subdirectories in the given directory."""
    # TODO: I think there is a better way to return this data. Rather than ['foo', 'subfoo'], I'd like to see something like:
    # {
    #   'foo': {
    #     'subfoo': {}
    #   }
    # }
    subdir_names = []
    for path, dirs, files in os.walk(directory_path):
        if any(dirs):
            subdir_names.extend(dirs)
        if not recursive:
            break
    return subdir_names


def directory_files_containing(
    directory_path: str, pattern: str, *, pattern_is_regex: bool = False, recursive: bool = True
) -> Dict[str, List[str]]:
    """Search for the given pattern in all files in the given directory_path."""
    from .files import file_name, file_search

    matching_files = {}
    file_paths = directory_file_paths(directory_path, recursive=recursive)

    for file_path in file_paths:
        search_results = file_search(file_path, pattern, pattern_is_regex=pattern_is_regex)
        if any(search_results):
            matching_files[file_path] = search_results
    return matching_files


def directory_file_paths_matching(directory_path: str, pattern: str, *, recursive: bool = True) -> List[str]:
    """Return the paths of all of the files in the given directory which match the pattern."""
    from .files import file_name_matches, file_name

    # TODO: consider consolidating this function into the directory_file_names_matching function and providing a flag to specify whether or not the user wants a file name or file path

    matching_file_paths = [
        file_path
        for file_path in directory_file_paths(directory_path, recursive=recursive)
        if file_name_matches(file_path, pattern) or pattern in file_path
    ]
    return matching_file_paths


def directory_file_names_matching(directory_path: str, pattern: str, *, recursive: bool = True) -> List[str]:
    """Return the names of all of the files in the given directory which match the pattern."""
    from .files import file_name_matches, file_name

    matching_file_names = [
        name
        for name in directory_file_names(directory_path, recursive=recursive)
        if file_name_matches(name, pattern) or pattern in name
    ]
    return matching_file_names


def directory_read_files_with_path_matching(
    directory_path: str, pattern: str, *, recursive: bool = True
) -> Iterable[Tuple[str, str]]:
    """Read all of the files in the given directory whose paths match the given pattern."""
    from .files import file_read

    matching_file_paths = directory_file_paths_matching(directory_path, pattern, recursive=recursive)

    for file_path in matching_file_paths:
        yield file_path, file_read(file_path)


def temp_dir_create(**kwargs) -> str:
    """Create a temporary directory."""
    return tempfile.TemporaryDirectory(**kwargs)



import fnmatch
import os
import posixpath
import ntpath
import shutil
import sys
import tempfile
from typing import Any, Dict, Union, List

from .atomic_writes import atomic_write


def _file_active_action(file_path: str, base_mode: str, file_contents: Any):
    """Perform an active action (write or append) with the given file contents on the given file."""
    if isinstance(file_contents, str):
        length_of_content = _file_action(file_path, '{}+'.format(base_mode), 'write', file_contents)
    elif isinstance(file_contents, bytes):
        length_of_content = _file_action(file_path, '{}b+'.format(base_mode), 'write', file_contents)
    else:
        print(f'Converting file contents of type {type(file_contents)} to string')
        length_of_content = _file_active_action(file_path, base_mode, str(file_contents))

    if length_of_content >= 0:
        result = True
    else:
        # I'm not sure if there are any cases in which this would occur, but just have it here for safety
        result = False

    return result


def _file_action(file_path, mode='r', command='read', contents=None):
    if 'w' in mode:
        with atomic_write(file_path, mode=mode) as f:
            return eval(f'f.{command}(contents)')
    else:
        with open(file_path, mode) as f:
            if contents:
                return eval(f'f.{command}(contents)')
            else:
                return eval(f'f.{command}()')


def is_file(path: str) -> bool:
    """Determine if the given path is a file."""
    return os.path.isfile(path)


def file_read(file_path: str) -> str:
    file_text = _file_action(file_path, 'r', 'read')
    return file_text


def file_read_bytes(file_path: str) -> bytes:
    file_text = _file_action(file_path, 'rb', 'read')
    return file_text


def file_write(file_path: str, file_contents: Any) -> bool:
    """Write the given content to the file at the given path (including a file name)."""
    result = _file_active_action(file_path, 'w', file_contents)
    return result


def file_append(file_path: str, file_contents: Any) -> bool:
    """Append the given content to the file at the given path (including a file name)."""
    result = _file_active_action(file_path, 'a', file_contents)
    return result


def file_move(starting_path: str, destination_path: str):
    """Move the file from the starting path to the destination path."""
    # TODO: this function could also be called "file_rename" - add a tag so that this function shows up if someone searches for "rename"
    shutil.move(starting_path, destination_path)


def file_copy(starting_path: str, destination_path: str, *, preserve_metadata: bool = False):
    """Copy the file from the starting_path to the destination path."""
    if preserve_metadata:
        shutil.copy2(starting_path, destination_path)
    else:
        shutil.copy(starting_path, destination_path)


def file_delete(file_path: str):
    """Delete the given file."""
    os.remove(file_path)


def file_owner_name(file_path: str) -> str:
    """Find the owner of the file at the given path."""
    # TODO: I believe this function only works for unix systems; update it (or write a corresponding function) to work on windows too
    from pwd import getpwuid

    file_owner_uid = os.stat(file_path).st_uid
    owner_username = getpwuid(file_owner_uid).pw_name
    return owner_username


# TODO: this function could also be called `fileChown` or `directory...`
# TODO: write a test for this function
def file_change_owner(file_path: str):
    """Change the ownership of the given file."""
    shutil.chown(file_path)


def file_ssdeep(file_path: str) -> str:
    """Find the ssdeep fuzzy hash of the file."""
    from d8s_hashes import ssdeep

    return ssdeep(file_read_bytes(file_path))


def file_md5(file_path: str) -> str:
    """Find the md5 hash of the given file."""
    from d8s_hashes import md5

    return md5(file_read_bytes(file_path))


def file_sha1(file_path: str) -> str:
    """Find the sha1 hash of the given file."""
    from d8s_hashes import sha1

    return sha1(file_read_bytes(file_path))


def file_sha256(file_path: str) -> str:
    """Find the sha256 hash of the given file."""
    from d8s_hashes import sha256

    return sha256(file_read_bytes(file_path))


def file_sha512(file_path: str) -> str:
    """Find the sha512 hash of the given file."""
    from d8s_hashes import sha512

    return sha512(file_read_bytes(file_path))


def file_name_escape(file_name: str) -> str:
    """Escape the name of a file so that it can be used as a file name in a file path."""
    import urllib.parse as urlparse

    # TODO: I should probably make an 'unescape' file

    url_encoded_file_name = urlparse.quote_plus(file_name)
    return url_encoded_file_name


def file_name(file_path: str) -> str:
    """Find the file name from the given file path."""
    return os.path.basename(file_path)


def file_name_windows(windows_file_path: str) -> str:
    """Find the file name from the given windows_file_path."""
    return ntpath.basename(windows_file_path)


def file_name_unix(unix_file_path: str) -> str:
    """Find the file name from the given unix_file_path."""
    return posixpath.basename(unix_file_path)


def file_size(file_path: str) -> int:
    """Find the file size."""
    return os.stat(file_path).st_size


def file_directory(file_path: str) -> str:
    """Return the directory in which the given file resides."""
    return file_path.replace(file_name(file_path), '')


def file_details(file_path: str) -> Dict[str, Union[str, int]]:
    """Get file hashes and file size for the given file."""
    file_details = {
        'md5': file_md5(file_path),
        'sha1': file_sha1(file_path),
        'sha256': file_sha256(file_path),
        'ssdeep': file_ssdeep(file_path),
        'size': file_size(file_path),
    }
    return file_details


def file_exists(file_path: str) -> bool:
    """Check if the file exists."""
    return os.access(file_path, os.F_OK)


def file_is_readable(file_path: str) -> bool:
    """Check if the file is readable."""
    return os.access(file_path, os.R_OK)


def file_is_writable(file_path: str) -> bool:
    """Check if the file is writable."""
    return os.access(file_path, os.W_OK)


def file_is_executable(file_path: str) -> bool:
    """Check if the file is executable."""
    return os.access(file_path, os.X_OK)


def file_contains(file_path: str, pattern: str, *, pattern_is_regex: bool = False) -> bool:
    """Return whether or not the file contains the given pattern."""
    result = file_search(file_path, pattern, pattern_is_regex=pattern_is_regex)
    return any(result)


def file_search(file_path: str, pattern: str, *, pattern_is_regex: bool = False) -> List[str]:
    """Search for the given pattern in the file."""
    import re

    file_text = file_read(file_path)
    if pattern_is_regex:
        return re.findall(pattern, file_text)
    else:
        return [pattern] * file_text.count(pattern)


def file_name_matches(file_path: str, pattern: str) -> bool:
    """Return whether or not the file name contains the given pattern."""
    name = file_name(file_path)
    return fnmatch.fnmatch(name, pattern)


def temp_file_create(**kwargs) -> str:
    """Create a temporary file."""
    return tempfile.TemporaryFile(**kwargs)
